package client;

import common.ChatMessageListener;
import common.CombinedSymmetricCryptoTransform;
import common.Logger;
import common.ROT13;
import common.SwapCrypt;
import common.SymmetricCryptoTransform;
import common.TextMessage;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.net.Socket;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;

/**
 * simple chat client
 */
public class Client implements Runnable {

    private final Socket socket;

    private final ObjectInputStream inputStream;

    private final ObjectOutputStream outputStream;

    private volatile boolean keepRunning;

    /**
     * listener-list for the observer pattern
     */
    private final Collection listeners;

    /* FEATURE: Encryption */
    private final SymmetricCryptoTransform cryptoTransform;

    /* FEATURE: Color */
    private String color;

    public Client(String host, int port) throws IOException {
        /* FEATURE: Encryption */
        cryptoTransform = new CombinedSymmetricCryptoTransform(new SymmetricCryptoTransform[]{
                new ROT13(), new SwapCrypt() });
        System.out.println("Connecting to " + host + " (port " + port + ")...");
        socket = new Socket(host, port);
        outputStream = new ObjectOutputStream(socket.getOutputStream());
        inputStream = new ObjectInputStream(socket.getInputStream());
        keepRunning = true;
        listeners = new ArrayList();
    }

    public static void main(String[] args) {
        if (args.length != 2) {
            throw new RuntimeException("Syntax: ChatClient <host> <port>");
        }

        try {
            Client client = new Client(args[0], Integer.parseInt(args[1]));
            /* FEATURE: Logging */
            client.addMessageListener(new Logger("client_" + args[0] + '_' + args[1]));

            Gui gui = new Gui("Chat " + args[0] + ':' + args[1], client);
            gui.pack();

            gui.setVisible(true);
            new Thread(client).start();
        } catch (IOException ignored) {
            System.out.println("Could not connect to server");
        }
    }

    /**
     * main method. waits for incoming messages.
     */
    public void run() {
        try {
            /* FEATURE: Color */
            fireNewMessage(new TextMessage("[color]",
                                           "Use \"/color [newColor]\" to set the color for "
                                           + "all messages or \"/color\" to reset the color"));
            while (keepRunning) {
                Object msg = inputStream.readObject();
                handleIncomingMessage(msg);
            }
        } catch (IOException ignored) {
            fireNewMessage(new TextMessage(TextMessage.SYSTEM_SENDER_NAME,
                                           "Server not available"));
        } catch (ClassNotFoundException ex) {
            ex.printStackTrace();
        } finally {
            try {
                socket.close();
                inputStream.close();
                outputStream.close();
            } catch (IOException ignored) {
            }
        }
    }

    /**
     * decides what to do with incoming messages
     *
     * @param msg the message (Object) received from the sockets
     */
    private void handleIncomingMessage(Object msg) {
        if (msg instanceof TextMessage) {
            TextMessage message = (TextMessage) msg;

            /* FEATURE: Encryption */
            // system messages are not encrypted
            if (!message.getSender().equals(TextMessage.SYSTEM_SENDER_NAME)) {
                String decryptedContent = cryptoTransform.transform(message.getContent());
                message = new TextMessage(message.getSender(), decryptedContent,
                                          message.getColor());
            }

            fireNewMessage(message);
        }
    }

    public void send(String message) {
        /* FEATURE: Color */
        if (message.startsWith("/color")) {
            if (message.length() == 6) {
                color = null;
                fireNewMessage(new TextMessage("[color]", "Color reset"));
            } else {
                color = message.substring(7);
                fireNewMessage(new TextMessage("[color]", "Color set to \"" + color + '"'));
            }
            return;
        }

        /* FEATURE: Encryption */
        message = cryptoTransform.transform(message);

        /* FEATURE: Color */
        TextMessage messageObject = new TextMessage(null, message, color);
        try {
            outputStream.writeObject(messageObject);
            outputStream.flush();
        } catch (IOException ignored) {
            fireNewMessage(new TextMessage(TextMessage.SYSTEM_SENDER_NAME,
                                           "Server not available"));
            stop();
        }
    }

    /**
     * addListener method for the observer pattern
     */
    public void addMessageListener(ChatMessageListener listener) {
        listeners.add(listener);
    }

    /**
     * removeListener method for the observer pattern
     */
    public void removeMessageListener(ChatMessageListener listener) {
        listeners.remove(listener);
    }

    /**
     * fire event method for the observer pattern
     */
    private void fireNewMessage(TextMessage message) {
        Iterator iterator = listeners.iterator();
        while (iterator.hasNext()) {
            Object next = iterator.next();
            if (next instanceof ChatMessageListener) {
                ((ChatMessageListener) next).newChatMessage(message);
            }
        }
    }

    public void stop() {
        keepRunning = false;
    }
}
