package server;

import common.ChatMessageListener;
import common.TextMessage;

import java.io.EOFException;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.net.Socket;
import java.net.SocketException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;

/**
 * class for an individual connection to a client. allows to send messages to
 * this client and handles incoming messages.
 */
public class Connection implements Runnable {

    private final Socket socket;

    private final Server server;

    private final ObjectInputStream inputStream;

    private final ObjectOutputStream outputStream;

    private final Collection listeners;

    public Connection(Socket socket, Server server) throws IOException {
        this.socket = socket;
        this.server = server;
        inputStream = new ObjectInputStream(socket.getInputStream());
        outputStream = new ObjectOutputStream(socket.getOutputStream());
        listeners = new ArrayList();
    }

    /**
     * waits for incoming messages from the socket
     */
    public void run() {
        String clientName = socket.getInetAddress().toString();
        try {
            server.broadcast(new TextMessage(TextMessage.SYSTEM_SENDER_NAME,
                                             clientName + " has joined."));
            Object msg;
            while ((msg = inputStream.readObject()) != null) {
                if (msg instanceof TextMessage) {
                    TextMessage message = (TextMessage) msg;
                    handleIncomingMessage(new TextMessage(clientName, message.getContent(),
                                                          message.getColor()));
                }
            }
        } catch (SocketException ignored) {
        } catch (EOFException ignored) {
        } catch (ClassNotFoundException e) {
            e.printStackTrace();
        } catch (IOException e) {
            e.printStackTrace();
        } finally {
            server.removeConnection(this);
            server.broadcast(new TextMessage(TextMessage.SYSTEM_SENDER_NAME,
                                             clientName + " has left."));
            try {
                socket.close();
            } catch (IOException ex) {
                ex.printStackTrace();
            }
        }
    }

    /**
     * decides what to do with incoming messages
     *
     * @param message the received message
     */
    private void handleIncomingMessage(TextMessage message) {
        fireNewMessage(message);
        server.broadcast(message);
    }

    /**
     * sends a message to the client
     *
     * @param message the message to send to the client
     */
    public void send(TextMessage message) {
        try {
            synchronized (outputStream) {
                outputStream.writeObject(message);
            }
            outputStream.flush();
        } catch (IOException ex) {
            ex.printStackTrace();
        }
    }

    /**
     * addListener method for the observer pattern
     */
    public void addLineListener(ChatMessageListener listener) {
        listeners.add(listener);
    }

    /**
     * removeListener method for the observer pattern
     */
    public void removeLineListener(ChatMessageListener listener) {
        listeners.remove(listener);
    }

    /**
     * fire event method for the observer pattern
     */
    private void fireNewMessage(TextMessage message) {
        Iterator iterator = listeners.iterator();
        while (iterator.hasNext()) {
            ChatMessageListener listener = (ChatMessageListener) iterator.next();
            listener.newChatMessage(message);
        }
    }

}
