package server;

import common.Logger;
import common.TextMessage;

import java.io.IOException;
import java.net.ServerSocket;
import java.net.Socket;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;

/**
 * server's main class. accepts incoming connections and allows broadcasting
 */
public class Server implements Runnable {

    private final int port;

    /**
     * list of all known connections to broadcast to
     */
    private final Collection connections;

    /* FEATURE: Logging */
    private final Logger logger;

    /**
     * awaits incoming connections and creates Connection objects accordingly.
     *
     * @param port port to listen on
     */
    public Server(int port) throws IOException {
        this.port = port;
        connections = new ArrayList();
        /* FEATURE: Logging */
        logger = new Logger("server_" + port);
    }

    public static void main(String[] args) {
        if (args.length != 1) {
            throw new RuntimeException("Syntax: ChatServer <port>");
        }
        try {
            Runnable server = new Server(Integer.parseInt(args[0]));
            server.run();
        } catch (IOException e) {
            System.out.println("Could not start server: " + e.getMessage());
        }
    }

    public void run() {
        ServerSocket server = null;
        try {
            server = new ServerSocket(port);
            while (true) {
                System.out.println("Waiting for Connections...");
                Socket client = server.accept();
                System.out.println("Accepted from " + client.getInetAddress());
                Connection newConnection = connectTo(client);
                new Thread(newConnection).start();
            }
        } catch (IOException e) {
            e.printStackTrace();
        } finally {
            if (server != null) {
                try {
                    server.close();
                } catch (IOException ignored) {
                }
            }
        }
    }

    /**
     * creates a new connection for a socket
     *
     * @param socket socket
     *
     * @return the Connection object that handles all further communication with
     * this socket
     */
    private Connection connectTo(Socket socket) {
        Connection connection = null;
        try {
            connection = new Connection(socket, this);

            /* FEATURE: Logging */
            connection.addLineListener(logger);

            connections.add(connection);
        } catch (IOException e) {
            e.printStackTrace();
        }
        return connection;
    }

    /**
     * send a message to all known connections
     *
     * @param message the message to broadcast
     */
    public void broadcast(TextMessage message) {
        synchronized (connections) {
            Iterator iterator = connections.iterator();
            while (iterator.hasNext()) {
                Object next = iterator.next();
                if (next instanceof Connection) {
                    ((Connection) next).send(message);
                }
            }
        }
    }

    /**
     * remove a connection so that broadcasts are no longer sent there.
     *
     * @param connection connection to remove
     */
    public void removeConnection(Connection connection) {
        connections.remove(connection);
    }
}
